unit UEngine;

{$mode delphi}

interface

uses
  Classes, SysUtils, Contnrs, Graphics, Controls, ExtCtrls, Math, DateUtils;

type
  TStationShape = (ssCircle, ssSquare, ssTriangle, ssStar, ssPlus, ssPentagon,
    ssDiamond);
  TStationShapeSet = set of TStationShape;
  TEngine = class;
  TMetroPassengers = class;
  TMetroLines = class;
  TMetroTrains = class;

  { TMetroStation }

  TMetroStation = class
  private
  public
    Engine: TEngine;
    Shape: TStationShape;
    Position: TPoint;
    Passengers: TMetroPassengers;
    Lines: TMetroLines;
    ShapeDistance: array[TStationShape] of Integer;
    function GetBestStationForShape(Shape: TStationShape; Check: TMetroStation;
      CurrentStationIndex: Integer): Boolean;
    constructor Create;
    destructor Destroy; override;
  end;

  { TMetroStations }

  TMetroStations = class(TObjectList)
    Engine: TEngine;
    function AddNew: TMetroStation;
  end;

  TTrackPoint = record
    StationIndex: Integer;
    Point: TPoint;
  end;

  { TMetroLine }

  TMetroLine = class
    Engine: TEngine;
    Color: TColor;
    Stations: TMetroStations;
    Trains: TMetroTrains;
    TrackPoints: array of TTrackPoint;
    procedure ConnectStation(Station: TMetroStation);
    procedure DisconnectStation(Station: TMetroStation);
    procedure AddTrack(P1, P2: TPoint; StationIndex: Integer);
    function GetTrackLength: Integer;
    function GetStationTrackPos(StationIndex: Integer): Integer;
    constructor Create;
    destructor Destroy; override;
    function IsCircular: Boolean;
  end;

  { TMetroLines }

  TMetroLines = class(TObjectList)
    Engine: TEngine;
    function AddNew: TMetroLine;
    function SearchByColor(Color: TColor): TMetroLine;
  end;

  { TMetroTrain }

  TMetroTrain = class
  private
    LastPosDelta: Integer;
    function GetTargetStation: TMetroStation;
  public
    Passengers: TMetroPassengers;
    Line: TMetroLine;
    TrackPos: Integer;
    Direction: Integer;
    InStation: Boolean;
    StationStopTime: TDateTime;
    TargetStationIndex: Integer;
    function GetPosition: TPoint;
    constructor Create;
    destructor Destroy; override;
    property TargetStation: TMetroStation read GetTargetStation;
  end;

  { TMetroTrains }

  TMetroTrains = class(TObjectList)
    function GetUnusedTrain: TMetroTrain;
    function GetUnusedCount: Integer;
    function AddNew: TMetroTrain;
  end;

  TMetroPassenger = class
    Engine: TEngine;
    Shape: TStationShape;
    Station: TMetroStation;
    Train: TMetroTrain;
  end;

  { TMetroPassengers }

  TMetroPassengers = class(TObjectList)
    Engine: TEngine;
    function AddNew: TMetroPassenger;
  end;

  TMap = class
    Size: TPoint;
  end;

  TView = class
    Size: TPoint;
  end;

  TGameState = (gsNotStarted, gsRunning, gsPaused, gsGameOver);

  { TEngine }

  TEngine = class
  private
    LastMousePos: TPoint;
    LastSelectedStation: TMetroStation;
    MouseHold: Boolean;
    LastNewStationTime: TDateTime;
    LastNewPassengerTime: TDateTime;
    LastNewWeekTime: TDateTime;
    LastNewShapeTime: TDateTime;
    ImagePassenger: TImage;
    ImageLocomotive: TImage;
    function GetExistStationShapes: TStationShapeSet;
    function GetStationOnPos(Pos: TPoint): TMetroStation;
    function GetLineOnPos(Pos: TPoint): TMetroLine;
    procedure DrawLine(Canvas: TCanvas; Pos: TPoint);
    procedure DrawShape(Canvas: TCanvas; Position: TPoint; Shape: TStationShape;
      Size: Integer);
    procedure DrawClock(Canvas: TCanvas);
    procedure ComputeShapeDistance;
    procedure ComputeShapeDistanceStation(Station: TMetroStation;
      UpdatedShape: TStationShape; Distance: Integer);
    function GetTime: TDateTime;
  public
    Passengers: TMetroPassengers;
    Stations: TMetroStations;
    Lines: TMetroLines;
    Trains: TMetroTrains;
    ShapeCount: Integer;
    Map: TMap;
    View: TView;
    SelectedLine: TMetroLine;
    SelectedStation: TMetroStation;
    ServedPassengerCount: Integer;
    State: TGameState;
    StartTime: TDateTime;
    procedure Tick;
    procedure MouseMove(Position: TPoint);
    procedure MouseUp(Button: TMouseButton; Position: TPoint);
    procedure MouseDown(Button: TMouseButton; Position: TPoint);
    procedure Reset;
    constructor Create;
    destructor Destroy; override;
    procedure Paint(Canvas: TCanvas);
    property Time: TDateTime read GetTime;
  end;

const
  clDarkYellow = TColor($00eeee);
  LineColors: array[0..5] of TColor = (clBlue, clRed, clDarkYellow, clGreen,
    clPurple, clGray);
  StationSize = 30;
  PassengerSize = 15;
  TrainSize = 40;
  LineColorsDist = 50;
  TrainSpeed = 3;
  ImagePassengerName = 'Images/Passenger.png';
  ImageLocomotiveName = 'Images/Locomotive.png';
  TrainPassengerCount = 6;
  StationMinDistance = 100;
  StationMaxDistance = 300;
  MaxWaitingPassengers = 10;

implementation

uses
  UGeometric;

{ TMetroPassengers }

function TMetroPassengers.AddNew: TMetroPassenger;
begin
  Result := TMetroPassenger.Create;
  Result.Engine := Engine;
  Result.Shape := TStationShape(Random(Integer(Engine.ShapeCount)));
  Add(Result);
end;

{ TMetroTrains }

function TMetroTrains.GetUnusedTrain: TMetroTrain;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Assigned(TMetroTrain(Items[I]).Line)) do Inc(I);
  if I < Count then Result := TMetroTrain(Items[I])
    else Result := nil;
end;

function TMetroTrains.GetUnusedCount: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to Count - 1 do
    if not Assigned(TMetroTrain(Items[I]).Line) then Inc(Result);
end;

function TMetroTrains.AddNew: TMetroTrain;
begin
  Result := TMetroTrain.Create;
  Add(Result);
end;

{ TMetroStations }

function TMetroStations.AddNew: TMetroStation;
var
  D: Integer;
  MinD: Integer;
  I: Integer;
  Pass: Integer;
  Angle: Double;
  L: Integer;
const
  Step = 20;
begin
  Result := TMetroStation.Create;
  Result.Engine := Engine;
  Angle := Random * 2 * Pi;
  // Ensure minimum distance between stations
  Pass := 0;
  L := Step;
  repeat
    Result.Position := Point(Trunc(Engine.Map.Size.X / 2 + Cos(Angle) * L),
      Trunc(Engine.Map.Size.Y / 2 + Sin(Angle) * L));
    MinD := High(Integer);
    for I := 0 to Engine.Stations.Count - 1 do begin
      D := Distance(TMetroStation(Engine.Stations[I]).Position, Result.Position);
      if D < MinD then MinD := D;
    end;
    Inc(Pass);
    L := L + StationMinDistance div 2;
  until (MinD > StationMinDistance) or
    (Pass > 1000) or (Engine.Stations.Count = 0);
  Result.Shape := TStationShape(Random(Integer(Engine.ShapeCount)));
  Add(Result);
  Engine.ComputeShapeDistance;
end;

{ TMetroLines }

function TMetroLines.AddNew: TMetroLine;
begin
  Result := TMetroLine.Create;
  Result.Color := LineColors[Count];
  Result.Engine := Engine;
  Add(Result);
end;

function TMetroLines.SearchByColor(Color: TColor): TMetroLine;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TMetroLine(Items[I]).Color <> Color) do Inc(I);
  if I < Count then Result := TMetroLine(Items[I])
    else Result := nil;
end;

{ TMetroLine }

procedure TMetroLine.ConnectStation(Station: TMetroStation);
var
  Train: TMetroTrain;
  StationIndex: Integer;
begin
  Stations.Add(Station);
  Station.Lines.Add(Self);
  StationIndex := Stations.Count - 1;
  if Stations.Count = 1 then begin
    SetLength(TrackPoints, Length(TrackPoints) + 1);
    TrackPoints[High(TrackPoints)].StationIndex := StationIndex;
    TrackPoints[High(TrackPoints)].Point := Station.Position;
  end else begin
     if Stations.Count > 1 then
       AddTrack(TMetroStation(Stations[Stations.Count - 2]).Position, Station.Position, StationIndex);
       // Place one train if at least two stations present
       if (Stations.Count = 2) then begin
         Train := Engine.Trains.GetUnusedTrain;
         if Assigned(Train) then begin
           Train.Line := Self;
           Train.TargetStationIndex := 0;
           Trains.Add(Train);
         end;
       end;
  end;
  Engine.ComputeShapeDistance;
end;

procedure TMetroLine.DisconnectStation(Station: TMetroStation);
var
  I: Integer;
  StationIndex: Integer;
begin
  if (Stations.Count > 0) and (Stations.Last = Station) then begin
    StationIndex := Stations.Count - 1;
    Stations.Delete(StationIndex);
    Station.Lines.Remove(Self);
    if Stations.Count > 0 then begin
      while (Length(TrackPoints) > 0) and (not ComparePoint(TrackPoints[High(TrackPoints)].Point,
        TMetroStation(Stations.Last).Position)) do
        SetLength(TrackPoints, Length(TrackPoints) - 1);
    end else SetLength(TrackPoints, 0);
    // Remove trains if less then two stations
    if Stations.Count < 2 then
      for I := Trains.Count - 1 downto 0 do begin
        TMetroTrain(Trains[I]).Line := nil;
        Trains.Delete(I);
      end;
  end;
  Engine.ComputeShapeDistance;
end;

procedure TMetroLine.AddTrack(P1, P2: TPoint; StationIndex: Integer);
var
  Delta: TPoint;
begin
  Delta := Point(P2.X - P1.X, P2.Y - P1.Y);
  if Abs(Delta.X) > Abs(Delta.Y) then begin
    SetLength(TrackPoints, Length(TrackPoints) + 1);
    TrackPoints[High(TrackPoints)].Point := Point(P2.X - Sign(Delta.X) * Abs(Delta.Y), P1.Y);
    TrackPoints[High(TrackPoints)].StationIndex := StationIndex - 1;
  end else begin
    SetLength(TrackPoints, Length(TrackPoints) + 1);
    TrackPoints[High(TrackPoints)].Point := Point(P1.X, P2.Y - Sign(Delta.Y) * Abs(Delta.X));
    TrackPoints[High(TrackPoints)].StationIndex := StationIndex - 1;
  end;
  SetLength(TrackPoints, Length(TrackPoints) + 1);
  TrackPoints[High(TrackPoints)].Point := P2;
  TrackPoints[High(TrackPoints)].StationIndex := StationIndex;
end;

function TMetroLine.GetTrackLength: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to High(TrackPoints) do
  if I > 0 then
    Result := Result + Distance(TrackPoints[I].Point, TrackPoints[I - 1].Point);
end;

function TMetroLine.GetStationTrackPos(StationIndex: Integer): Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to High(TrackPoints) do begin
    if I > 0 then
      Result := Result + Distance(TrackPoints[I].Point, TrackPoints[I - 1].Point);
    if TrackPoints[I].StationIndex = StationIndex then Break;
  end;
end;

constructor TMetroLine.Create;
begin
  Stations := TMetroStations.Create;
  Stations.OwnsObjects := False;
  Trains := TMetroTrains.Create;
  Trains.OwnsObjects := False;
end;

destructor TMetroLine.Destroy;
begin
  Trains.Free;
  Stations.Free;
  inherited Destroy;
end;

function TMetroLine.IsCircular: Boolean;
begin
  Result := False;
  if Stations.Count >= 2 then
    Result := (Stations.Last = Stations.First);
end;

{ TMetroTrain }

function TMetroTrain.GetTargetStation: TMetroStation;
begin
  Result := nil;
  if Assigned(Line) then
  if (TargetStationIndex >= 0) and (TargetStationIndex < Line.Stations.Count) then
    Result := TMetroStation(Line.Stations[TargetStationIndex]);
end;

function TMetroTrain.GetPosition: TPoint;
var
  I: Integer;
  R: TPoint;
  D: Integer;
  Sum: Integer;
begin
  if Assigned(Line) then
  with Line do begin
    Sum := 0;
    if Length(TrackPoints) > 1 then
    for I := 1 to High(TrackPoints) do begin
      D := Distance(TrackPoints[I].Point, TrackPoints[I - 1].Point);
      if (Sum + D) > Self.TrackPos then begin
        R := Point(TrackPoints[I].Point.X - TrackPoints[I - 1].Point.X,
          TrackPoints[I].Point.Y - TrackPoints[I - 1].Point.Y);
        Result := Point(Trunc(TrackPoints[I - 1].Point.X + R.X * (TrackPos - Sum) / D),
          Trunc(TrackPoints[I - 1].Point.Y + R.Y * (TrackPos - Sum) / D));
        Exit;
      end else Sum := Sum + D;
    end;
    Result := TrackPoints[High(TrackPoints)].Point;
  end else Result := Point(0, 0);
end;

constructor TMetroTrain.Create;
begin
  Passengers := TMetroPassengers.Create;
  Passengers.OwnsObjects := False;
  Direction := 1;
  Line := nil;
end;

destructor TMetroTrain.Destroy;
begin
  Passengers.Free;
  inherited Destroy;
end;

{ TMetroStation }

function TMetroStation.GetBestStationForShape(Shape: TStationShape;
  Check: TMetroStation; CurrentStationIndex: Integer): Boolean;
var
  I: Integer;
  T: Integer;
  Distance: Integer;
  StationIndex: Integer;
  DirectionUp: Boolean;
  DirectionDown: Boolean;
  NextStationUp: TMetroStation;
  NextStationDown: TMetroStation;
begin
  Distance := High(Integer);
  for I := 0 to Lines.Count - 1 do
  with TMetroLine(Lines[I]) do begin
    StationIndex := CurrentStationIndex;
    if IsCircular then begin
      DirectionUp := False;
      DirectionDown := False;
      for T := 0 to Trains.Count - 1 do begin
        if TMetroTrain(Trains[T]).Direction = 1 then DirectionUp := True;
        if TMetroTrain(Trains[T]).Direction = -1 then DirectionDown := True;
      end;
      if StationIndex = 0 then
        NextStationDown := TMetroStation(Stations[Stations.Count - 2])
      else
      if StationIndex > 0 then
        NextStationDown := TMetroStation(Stations[StationIndex - 1]);

      if (StationIndex >= 0) and (StationIndex = Stations.Count - 1) then
        NextStationUp := TMetroStation(Stations[1])
      else
      if (StationIndex >= 0) and (StationIndex < Stations.Count - 1) then
        NextStationUp := TMetroStation(Stations[StationIndex + 1]);
    end else begin
      if StationIndex > 0 then begin
        DirectionDown := True;
        NextStationDown := TMetroStation(Stations[StationIndex - 1])
      end else DirectionDown := False;
      if (StationIndex >= 0) and (StationIndex < Stations.Count - 1) then begin
        DirectionUp := True;
        NextStationUp := TMetroStation(Stations[StationIndex + 1]);
      end else DirectionUp := False;
    end;
    if DirectionDown and (NextStationDown.ShapeDistance[Shape] <> -1) and
      (NextStationDown.ShapeDistance[Shape] < Distance) then begin
      Distance := NextStationDown.ShapeDistance[Shape];
    end;
    if DirectionUp and (NextStationUp.ShapeDistance[Shape] <> -1) and
      (NextStationUp.ShapeDistance[Shape] < Distance) then begin
      Distance := NextStationUp.ShapeDistance[Shape];
    end;
  end;
  Result := Check.ShapeDistance[Shape] = Distance;
end;

constructor TMetroStation.Create;
begin
  Passengers := TMetroPassengers.Create;
  Passengers.OwnsObjects := False;
  Lines := TMetroLines.Create;
  Lines.OwnsObjects := False;
end;

destructor TMetroStation.Destroy;
begin
  Lines.Free;
  Passengers.Free;
  inherited Destroy;
end;

{ TEngine }

function TEngine.GetExistStationShapes: TStationShapeSet;
var
  I: Integer;
begin
  Result := [];
  for I := 0 to Stations.Count - 1 do
  with TMetroStation(Stations[I]) do begin
    Result := Result + [Shape];
  end;
end;

function TEngine.GetStationOnPos(Pos: TPoint): TMetroStation;
var
  I: Integer;
const
  ClickDistance = 30;
begin
  I := 0;
  while (I < Stations.Count) and (Distance(TMetroStation(Stations[I]).Position, Pos) > ClickDistance) do Inc(I);
  if I < Stations.Count then Result := TMetroStation(Stations[I])
    else Result := nil;
end;

function TEngine.GetLineOnPos(Pos: TPoint): TMetroLine;
var
  I: Integer;
  T: Integer;
  D: Integer;
const
  Distance = 30;
begin
  Result := nil;
  I := 0;
  while (I < Lines.Count) do
  with TMetroLine(Lines[I]) do begin
    for T := 1 to High(TrackPoints) do begin
      D := PointToLineDistance(Pos, TrackPoints[T - 1].Point, TrackPoints[T].Point);
      if D < Distance then begin
        Result := TMetroLine(Lines[I]);
        Break;
      end;
    end;
    if Assigned(Result) then Break;
    Inc(I);
  end;
end;

procedure TEngine.DrawLine(Canvas: TCanvas; Pos: TPoint);
var
  Delta: TPoint;
begin
  Delta := Point(Pos.X - Canvas.PenPos.X, Pos.Y - Canvas.PenPos.Y);
  if Abs(Delta.X) > Abs(Delta.Y) then begin
    Canvas.LineTo(Pos.X - Sign(Delta.X) * Abs(Delta.Y), Canvas.PenPos.Y);
  end else begin
    Canvas.LineTo(Canvas.PenPos.X, Pos.Y - Sign(Delta.Y) * Abs(Delta.X));
  end;
  Canvas.LineTo(Pos.X, Pos.Y);
end;

procedure TEngine.DrawShape(Canvas: TCanvas; Position: TPoint; Shape: TStationShape;
  Size: Integer);
var
  Points: array of TPoint;
  I: Integer;
  Angle: Double;
begin
  case Shape of
    ssSquare: Canvas.Rectangle(
      Position.X - Size div 2, Position.Y - Size div 2,
      Position.X + Size div 2, Position.Y + Size div 2);
    ssCircle: Canvas.Ellipse(
      Position.X - Size div 2, Position.Y - Size div 2,
      Position.X + Size div 2, Position.Y + Size div 2);
    ssTriangle: begin
      SetLength(Points, 3);
      Points[0] := Point(Position.X, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y + Size div 2);
      Points[2] := Point(Position.X - Size div 2, Position.Y + Size div 2);
      Canvas.Polygon(Points);
    end;
    ssStar: begin
      SetLength(Points, 10);
      for I := 0 to 9 do begin
        Angle := I / 10 * 2 * Pi - Pi / 2;
        if (I mod 2) = 0 then
        Points[I] := Point(Round(Position.X + Cos(Angle) * Size / 2),
          Round(Position.Y + Sin(Angle) * Size / 2))
        else
        Points[I] := Point(Round(Position.X + Cos(Angle) * Size / 5),
          Round(Position.Y + Sin(Angle) * Size / 5));
      end;
      Canvas.Polygon(Points);
    end;
    ssPlus: begin
      SetLength(Points, 12);
      Points[0] := Point(Position.X + Size div 6, Position.Y - Size div 6);
      Points[1] := Point(Position.X + Size div 2, Position.Y - Size div 6);
      Points[2] := Point(Position.X + Size div 2, Position.Y + Size div 6);
      Points[3] := Point(Position.X + Size div 6, Position.Y + Size div 6);
      Points[4] := Point(Position.X + Size div 6, Position.Y + Size div 2);
      Points[5] := Point(Position.X - Size div 6, Position.Y + Size div 2);
      Points[6] := Point(Position.X - Size div 6, Position.Y + Size div 6);
      Points[7] := Point(Position.X - Size div 2, Position.Y + Size div 6);
      Points[8] := Point(Position.X - Size div 2, Position.Y - Size div 6);
      Points[9] := Point(Position.X - Size div 6, Position.Y - Size div 6);
      Points[10] := Point(Position.X - Size div 6, Position.Y - Size div 2);
      Points[11] := Point(Position.X + Size div 6, Position.Y - Size div 2);
      Canvas.Polygon(Points);
    end;
    ssPentagon: begin
      SetLength(Points, 5);
      for I := 0 to 4 do begin
        Angle := I / 5 * 2 * Pi - Pi / 2;
        Points[I] := Point(Round(Position.X + Cos(Angle) * Size / 2),
          Round(Position.Y + Sin(Angle) * Size / 2));
      end;
      Canvas.Polygon(Points);
    end;
    ssDiamond: begin
      SetLength(Points, 4);
      Points[0] := Point(Position.X, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y);
      Points[2] := Point(Position.X, Position.Y + Size div 2);
      Points[3] := Point(Position.X - Size div 2, Position.Y);
      Canvas.Polygon(Points);
    end;
  end;
end;

procedure TEngine.ComputeShapeDistance;
var
  I: Integer;
  L: Integer;
  S: TStationShape;
begin
  // Reset all distances
  for I := 0 to Stations.Count - 1 do
  with TMetroStation(Stations[I]) do begin
    for S := Low(ShapeDistance) to High(ShapeDistance) do
      ShapeDistance[S] := -1;
  end;

  // Propagate shape distance for all stations
  // Distace 0 means that station is final target
  for I := 0 to Stations.Count - 1 do
  with TMetroStation(Stations[I]) do begin
    ComputeShapeDistanceStation(TMetroStation(Stations[I]), Shape, 0);
  end;
end;

procedure TEngine.ComputeShapeDistanceStation(Station: TMetroStation;
  UpdatedShape: TStationShape; Distance: Integer);
var
  I: Integer;
  T: Integer;
  StationIndex: Integer;
  DirectionDown: Boolean;
  DirectionUp: Boolean;
begin
  with Station do begin
    if (Distance < ShapeDistance[UpdatedShape]) or (ShapeDistance[UpdatedShape] = -1) then begin
      ShapeDistance[UpdatedShape] := Distance;
      // Do for all lines connected to station
      for I := 0 to Lines.Count - 1 do
      with TMetroLine(Lines[I]) do
        for StationIndex := 0 to Stations.Count - 1 do
        if TMetroStation(Stations[StationIndex]) = Station then begin
        if not IsCircular then begin
          // Update for all adjecent stations
          if StationIndex > 0 then
            ComputeShapeDistanceStation(TMetroStation(Stations[StationIndex - 1]), UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          if (StationIndex >= 0) and (StationIndex < Stations.Count - 1) then
            ComputeShapeDistanceStation(TMetroStation(Stations[StationIndex + 1]), UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
        end else begin
          // If circular then trains might go in single direction so passengers
          // waiting for oposite directions are wrong
          DirectionUp := False;
          DirectionDown := False;
          for T := 0 to Trains.Count - 1 do begin
            if TMetroTrain(Trains[T]).Direction = 1 then DirectionUp := True;
            if TMetroTrain(Trains[T]).Direction = -1 then DirectionDown := True;
          end;
          // Update for all adjecent stations
          if DirectionUp then begin
            if StationIndex = 0 then
              ComputeShapeDistanceStation(TMetroStation(Stations[Stations.Count - 2]), UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
            if StationIndex > 0 then
              ComputeShapeDistanceStation(TMetroStation(Stations[StationIndex - 1]), UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          end;
          if DirectionDown then begin
            if (StationIndex >= 0) and (StationIndex = Stations.Count - 1) then
              ComputeShapeDistanceStation(TMetroStation(Stations[1]), UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
            if (StationIndex >= 0) and (StationIndex < Stations.Count - 1) then
              ComputeShapeDistanceStation(TMetroStation(Stations[StationIndex + 1]), UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          end;
        end;
      end;
    end;
  end;
end;

function TEngine.GetTime: TDateTime;
begin
  Result := (Now - StartTime) / OneSecond * (60 * OneMinute);
end;

procedure TEngine.DrawClock(Canvas: TCanvas);
var
  ClockCenter: TPoint;
  Angle: Double;
  Text: string;
  I: Integer;
const
  ClockSize = 20;
begin
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Color := clBlack;
  Canvas.Pen.Width := 2;
  ClockCenter := Point(Canvas.Width - 30, 40);
  Angle := Time / (12 * OneHour) * 2 * Pi - Pi / 2;
  Canvas.EllipseC(ClockCenter.X, ClockCenter.Y, ClockSize, ClockSize);
  Canvas.Line(ClockCenter, Point(ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.8),
    ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.8)));
  Text := FormatDateTime('ddd', Time);
  Canvas.TextOut(ClockCenter.X - ClockSize - Canvas.TextWidth(Text) - 5, ClockCenter.Y -
    Canvas.TextWidth(Text) div 2, Text);
  for I := 0 to 12 do begin
    Angle := I / 12 * 2 * Pi;
    Canvas.Line(ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.8),
      ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.8),
      ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.9),
      ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.9));
  end;
end;

procedure TEngine.Tick;
const
  NewStationPeriod = OneSecond * 20;
  NewPassengerPeriod = OneSecond;
  NewPassengerProbability = 0.1;
var
  Passenger: TMetroPassenger;
  I: Integer;
  P: Integer;
  PosDelta: Integer;
  CurrentStation: TMetroStation;
  CurrentStationIndex: Integer;
begin
  if State = gsRunning then begin

  // Add new trains
  if (Now - LastNewWeekTime) > 7 then begin
    LastNewWeekTime := Now;
    Trains.AddNew;
    // TODO: Show notification screen with confirmation
  end;

  // Add new shape
  if (Now - LastNewShapeTime) > 10 then begin
    LastNewShapeTime := Now;
    if ShapeCount <= Integer(High(TStationShape)) then Inc(ShapeCount);
  end;

  // Add new stations
  if (Now - LastNewStationTime) > NewStationPeriod then begin
    LastNewStationTime := Now;
    Stations.AddNew;
  end;

  // Add new passengers
  if (Now - LastNewPassengerTime) > NewPassengerPeriod then begin
    LastNewPassengerTime := Now;
    for I := 0 to Stations.Count - 1 do
    with TMetroStation(Stations[I]) do
    if Random < NewPassengerProbability then begin
      Passenger := Self.Passengers.AddNew;
      Passenger.Station := TMetroStation(Stations[I]);
      Passengers.Add(Passenger);

      // Passenger is not allowed to have same shape
      while (Passenger.Shape = Passenger.Station.Shape) or not (Passenger.Shape in GetExistStationShapes) do
        Passenger.Shape := TStationShape((Integer(Passenger.Shape) + 1) mod Integer(ShapeCount));
    end;
  end;

  // Move trains
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do begin
    if Assigned(Line) then begin
      if InStation then begin
        if (Now - StationStopTime) > OneSecond then begin
          CurrentStationIndex := TargetStationIndex;
          if CurrentStationIndex >= Line.Stations.Count then
            CurrentStationIndex := Line.Stations.Count - 1;
          CurrentStation := nil;
          if Assigned(Line) then
            if (CurrentStationIndex >= 0) and (CurrentStationIndex < Line.Stations.Count) then
          CurrentStation := TMetroStation(Line.Stations[CurrentStationIndex]);

          // Choose next target station
          TargetStationIndex := TargetStationIndex + Direction;
          if TargetStationIndex < 0 then begin
            if Line.Stations.Last = Line.Stations.First then begin
              TargetStationIndex := Line.Stations.Count - 2;
              TrackPos := Line.GetStationTrackPos(Line.Stations.Count - 1);
            end else begin
              TargetStationIndex := 1;
              Direction := -Direction;
            end;
          end else if TargetStationIndex >= Line.Stations.Count then begin
            if Line.Stations.Last = Line.Stations.First then begin
              TargetStationIndex := 1;
              TrackPos := 0;
            end else begin
              TargetStationIndex := Line.Stations.Count - 2;
              Direction := -Direction;
            end;
          end;

          // Unload passengers in target station
          if Assigned(CurrentStation) and (CurrentStationIndex < Line.Stations.Count) then
          for P := Passengers.Count - 1 downto 0 do begin
            if TMetroPassenger(Passengers[P]).Shape = CurrentStation.Shape then begin
              Passenger := TMetroPassenger(Passengers[P]);
              Passengers.Delete(P);
              Self.Passengers.Remove(Passenger);
              Inc(ServedPassengerCount);
            end;
          end;
          // Unload passengers to change line
          if Assigned(CurrentStation) and (CurrentStationIndex < Line.Stations.Count) then
          for P := Passengers.Count - 1 downto 0 do begin
            if not CurrentStation.GetBestStationForShape(TMetroPassenger(Passengers[P]).Shape,
            TargetStation, CurrentStationIndex) then begin
              Passenger := TMetroPassenger(Passengers[P]);
              Passengers.Delete(P);
              CurrentStation.Passengers.Add(Passenger);
              Passenger.Station := CurrentStation;
            end;
          end;

          // Load new passengers
          if Assigned(CurrentStation) and (CurrentStationIndex < Line.Stations.Count) then
          for P := CurrentStation.Passengers.Count - 1 downto 0 do begin
            if (Passengers.Count < TrainPassengerCount) then begin
              Passenger := TMetroPassenger(CurrentStation.Passengers[P]);
              if CurrentStation.GetBestStationForShape(Passenger.Shape, TargetStation, CurrentStationIndex) then begin
                Passenger.Station := nil;
                CurrentStation.Passengers.Delete(P);
                Passengers.Add(Passenger);
                Passenger.Train := TMetroTrain(Trains[I]);
              end;
            end else Break; // No more space
          end;

          LastPosDelta := Abs(TrackPos - Line.GetStationTrackPos(TargetStationIndex));
          InStation := False;
        end;
      end else begin
      TrackPos := TrackPos + Direction * TrainSpeed;
      if TrackPos < 0 then begin
        if Line.Stations.First = Line.Stations.Last then TrackPos := Line.GetTrackLength
          else TrackPos := 0;
      end else
      if TrackPos > Line.GetTrackLength then begin
        if Line.Stations.First = Line.Stations.Last then TrackPos := 0
          else TrackPos := Line.GetTrackLength;
      end;
      PosDelta := Abs(TrackPos - Line.GetStationTrackPos(TargetStationIndex));
      if PosDelta >= LastPosDelta then begin
        // We are getting far from station, stop at station
        TrackPos := Line.GetStationTrackPos(TargetStationIndex);
        InStation := True;
        StationStopTime := Now;
      end;
      LastPosDelta := PosDelta;
      end;
    end;
  end;

  // Game over
  for I := 0 to Stations.Count - 1 do
  with TMetroStation(Stations[I]) do begin
    if Passengers.Count > MaxWaitingPassengers then State := gsGameOver;
  end;

  end;
end;

procedure TEngine.MouseMove(Position: TPoint);
var
  Station: TMetroStation;
begin
  LastMousePos := Position;
  if MouseHold then begin
    if Assigned(SelectedLine) then begin
      Station := GetStationOnPos(Position);
      if not Assigned(LastSelectedStation) and Assigned(Station) then begin
        if (SelectedLine.Stations.IndexOf(Station) = -1) then begin
          SelectedLine.ConnectStation(Station);
          SelectedStation := Station;
        end else
        if (SelectedLine.Stations.Count > 0) and (SelectedLine.Stations.Last = Station) then begin
          SelectedLine.DisconnectStation(Station);
          SelectedStation := TMetroStation(SelectedLine.Stations.Last);
        end else if (SelectedLine.Stations.Count > 0) and (SelectedLine.Stations.First = Station) then begin
          SelectedLine.ConnectStation(Station);
          SelectedStation := Station;
        end;
      end;
      LastSelectedStation := Station;
    end;
  end;
end;

procedure TEngine.MouseUp(Button: TMouseButton; Position: TPoint);
var
  Station: TMetroStation;
  Line: TMetroLine;
  I: Integer;
begin
  MouseHold := False;
  if Button = mbLeft then begin
    SelectedStation := nil;

    // Line color selection
    for I := 0 to Lines.Count - 1 do
      if Distance(Point(View.Size.X div 2 - Length(LineColors) div 2 * LineColorsDist + I * LineColorsDist,
        View.Size.Y - LineColorsDist), Position) < 20 then begin
          SelectedLine := TMetroLine(Lines[I]);
          Exit;
        end;
  end else
  if Button = mbRight then begin
    SelectedLine := nil;
  end;
end;

procedure TEngine.MouseDown(Button: TMouseButton; Position: TPoint);
var
  Station: TMetroStation;
  Line: TMetroLine;
  I: Integer;
begin
  if Button = mbLeft then begin
    MouseHold := True;
    LastSelectedStation := nil;

    // Station selection
    Station := GetStationOnPos(Position);
    if Assigned(Station) then begin
      SelectedStation := Station;
      if not Assigned(SelectedLine) then
      for I := 0 to Lines.Count - 1 do
      if TMetroLine(Lines[I]).Stations.Count = 0 then begin
        SelectedLine := TMetroLine(Lines[I]);
        Break;
      end;
    end;

    // Line selection
    Line := GetLineOnPos(Position);
    if Assigned(Line) then begin
      SelectedLine := Line;
      Exit;
    end;
  end;
end;

procedure TEngine.Reset;
var
  NewLine: TMetroLine;
  NewTrain: TMetroTrain;
  I: Integer;
  NewStation: TMetroStation;
begin
  Passengers.Clear;
  Lines.Clear;
  Stations.Clear;

  ShapeCount := 3;
  ServedPassengerCount := 0;

  // Start with 3 stations with each different shape
  NewStation := Stations.AddNew;
  NewStation.Shape := ssSquare;
  NewStation := Stations.AddNew;
  NewStation.Shape := ssCircle;
  NewStation := Stations.AddNew;
  NewStation.Shape := ssTriangle;

  for I := 0 to 2 do begin
    NewLine := Lines.AddNew;
    NewTrain := TMetroTrain.Create;
    Trains.Add(NewTrain);
  end;

  SelectedLine := TMetroLine(Lines[0]);
  LastNewStationTime := Now;
  LastNewPassengerTime := Now;
  LastNewWeekTime := Now;
  LastNewShapeTime := Now;
  State := gsRunning;
  StartTime := Now;
end;

constructor TEngine.Create;
begin
  Stations := TMetroStations.Create;
  Stations.Engine := Self;
  Lines := TMetroLines.Create;
  Lines.Engine := Self;
  Passengers := TMetroPassengers.Create;
  Passengers.Engine := Self;
  Map := TMap.Create;
  View := TView.Create;
  Trains := TMetroTrains.Create;
  ImagePassenger := TImage.Create(nil);
  ImageLocomotive := TImage.Create(nil);
  if FileExists(ImagePassengerName) then
    ImagePassenger.Picture.LoadFromFile(ImagePassengerName);
  if FileExists(ImageLocomotiveName) then
    ImageLocomotive.Picture.LoadFromFile(ImageLocomotiveName);
end;

destructor TEngine.Destroy;
begin
  Trains.Free;
  ImageLocomotive.Free;
  ImagePassenger.Free;
  View.Free;
  Map.Free;
  Passengers.Free;
  Stations.Free;
  Lines.Free;
  inherited Destroy;
end;

procedure TEngine.Paint(Canvas: TCanvas);
var
  I: Integer;
  S: Integer;
  Size: Integer;
  P: Integer;
  Pos: TPoint;
  Text: string;
  Angle: Double;
  EndPoint: TPoint;
const
  GameOverText = 'Game Over';
  GameOverReason = 'Overcrowding at this station has forced you to resign as metro manager.';
  GameOverStatistic = '%d passengers travelled on your metro over %d days.';
  EndStationLength = 50;
begin
  Canvas.Brush.Color := $eff0e0;
  Canvas.Brush.Style := bsSolid;
  Canvas.Clear;

  // Draw lines
  for I := 0 to Lines.Count - 1 do
  with TMetroLine(Lines[I]) do begin
    Canvas.Pen.Color := Color;
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := 15;
    if Length(TrackPoints) > 0 then Canvas.MoveTo(TrackPoints[0].Point);
    for S := 1 to High(TrackPoints) do begin
      Canvas.LineTo(TrackPoints[S].Point);
      if (S = High(TrackPoints)) then begin
        Angle := arctan2((TrackPoints[S].Point.Y - TrackPoints[S - 1].Point.Y), (TrackPoints[S].Point.X - TrackPoints[S - 1].Point.X));
        EndPoint := Point(Round(TrackPoints[S].Point.X + EndStationLength * Cos(Angle)),
          Round(TrackPoints[S].Point.Y + EndStationLength * Sin(Angle)));
        Canvas.LineTo(EndPoint);
        Canvas.MoveTo(Point(Round(EndPoint.X + Cos(Angle + Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle + Pi / 2) * EndStationLength / 3)));
        Canvas.LineTo(Point(Round(EndPoint.X + Cos(Angle - Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle - Pi / 2) * EndStationLength / 3)));
      end;
    end;
    if (Length(TrackPoints) > 1) then begin
      Angle := arctan2((TrackPoints[1].Point.Y - TrackPoints[0].Point.Y), (TrackPoints[1].Point.X - TrackPoints[0].Point.X));
      Canvas.MoveTo(TrackPoints[0].Point);
      EndPoint := Point(Round(TrackPoints[0].Point.X - EndStationLength * Cos(Angle)),
        Round(TrackPoints[0].Point.Y - EndStationLength * Sin(Angle)));
      Canvas.LineTo(EndPoint);
      Canvas.MoveTo(Point(Round(EndPoint.X - Cos(Angle + Pi / 2) * EndStationLength / 3),
        Round(EndPoint.Y - Sin(Angle + Pi / 2) * EndStationLength / 3)));
      Canvas.LineTo(Point(Round(EndPoint.X - Cos(Angle - Pi / 2) * EndStationLength / 3),
        Round(EndPoint.Y - Sin(Angle - Pi / 2) * EndStationLength / 3)));
    end;

    // Draw design line
    if (SelectedLine = TMetroLine(Lines[I])) and Assigned(SelectedStation) then begin
      Canvas.MoveTo(SelectedStation.Position);
      DrawLine(Canvas, LastMousePos);
    end;
  end;

  // Draw trains
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do begin
    if Assigned(Line) then begin
      Canvas.Brush.Color := Line.Color;
      Canvas.Brush.Style := bsSolid;
      Canvas.Pen.Style := psClear;
      Pos := GetPosition;
      Canvas.Rectangle(Pos.X - TrainSize div 2, Pos.Y - TrainSize div 2,
        Pos.X + TrainSize div 2, Pos.Y + TrainSize div 2);
      Canvas.Brush.Color := clWhite;
      for P := 0 to Passengers.Count - 1 do
      with TMetroPassenger(Passengers[P]) do begin
        DrawShape(Canvas, Point(Pos.X - Trunc(TrainSize div 3 * 1) + (P mod 3) * TrainSize div 3,
          Pos.Y - Trunc(TrainSize div 3 * 1) + (P div 3) * TrainSize div 3), Shape, TrainSize div 3);
      end;
    end;
  end;

  // Draw stations
  Canvas.Pen.Width := 5;
  for I := 0 to Stations.Count - 1 do
  with TMetroStation(Stations[I]) do begin
    Canvas.Pen.Style := psSolid;
    if Assigned(SelectedLine) and (Lines.IndexOf(SelectedLine) <> -1) then begin
      Canvas.Brush.Style := bsClear;
      Canvas.Pen.Color := SelectedLine.Color;
      DrawShape(Canvas, Position, Shape, StationSize + Canvas.Pen.Width + 6);
    end;
    Canvas.Brush.Color := clWhite;
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Color := clBlack;
    DrawShape(Canvas, Position, Shape, StationSize);

    // Draw passengers
    Canvas.Pen.Style := psClear;
    Canvas.Brush.Color := clBlack;
    for P := 0 to Passengers.Count - 1 do
    with TMetroPassenger(Passengers[P]) do begin
      DrawShape(Canvas, Point(Position.X + StationSize + P * (PassengerSize + 2), Position.Y - StationSize div 2),
        Shape, PassengerSize);
    end;

{
    Canvas.Brush.Style := bsClear;
    Text := '';
    for P := 0 to 2 do
      Text := Text + IntToStr(ShapeDistance[TStationShape(P)]) + ',';
    Canvas.TextOut(Position.X + StationSize div 2, Position.Y + StationSize div 2, Text);
    }
  end;

  // Line selection
  for I := 0 to High(LineColors) do begin
    if Assigned(Lines.SearchByColor(LineColors[I])) then begin
      Canvas.Brush.Color := LineColors[I];
      Size := 15;
    end else begin
      Canvas.Brush.Color := clSilver;
      Size := 5;
    end;
    Canvas.Pen.Color := clBlack;
    if Assigned(SelectedLine) and (SelectedLine.Color = LineColors[I]) then begin
      Canvas.Pen.Style := psSolid;
    end else begin
      Canvas.Pen.Style := psClear;
    end;

    Canvas.EllipseC(Canvas.Width div 2 - Length(LineColors) div 2 * LineColorsDist + I * LineColorsDist,
      Canvas.Height - LineColorsDist, Size, Size);
  end;

  // Interface
  Text := IntToStr(ServedPassengerCount);
  Canvas.Draw(Canvas.Width - 140, 20, ImagePassenger.Picture.Bitmap);
  Canvas.Brush.Style := bsClear;
  Canvas.TextOut(Canvas.Width - 146 - Canvas.TextWidth(Text), 25, Text);

  Text := IntToStr(Trains.GetUnusedCount);
  Canvas.Draw(Canvas.Width - 240, 20, ImageLocomotive.Picture.Bitmap);
  Canvas.Brush.Style := bsClear;
  Canvas.TextOut(Canvas.Width - 246 - Canvas.TextWidth(Text), 25, Text);

  DrawClock(Canvas);

  // Game over
  if State = gsGameOver then
  begin
    Canvas.Font.Size := 40;
    Canvas.Font.Color := clBlack;
    Canvas.TextOut((Canvas.Width - Canvas.TextWidth(GameOverText)) div 2, 100, GameOverText);
    Canvas.Font.Size := 14;
    Canvas.TextOut((Canvas.Width - Canvas.TextWidth(GameOverReason)) div 2, 160, GameOverReason);
    Text := Format(GameOverStatistic, [ServedPassengerCount, Trunc(Time)]);
    Canvas.TextOut((Canvas.Width - Canvas.TextWidth(Text)) div 2, 180, Text);
  end;
end;

end.

